/******************************** SBB **********************************
**
** Project      :       Multiple-Threaded Library (MTL)
**
** Author       :       Zhang Xiangyang
**
** Date         :       November 4, 1996
**
** Version      :       1.0
**              :       Copyright (C) 1996, Zhang Xiangyang
**                
** File name    :       pthread.c
**
** Description  :
**
**      This file is part of pthread Library, which provides the thread 
**      interface compatible with POSIX 1003.4a Reference.
**
**      Report problems and direct all questions to:
**       zhangx@sebb.bel.alcatel.be
**
********************************HISTORY*********************************
**   DATE    INITIALS	COMMENTS		
**   ----    --------	--------
** 04/11/96  XZHANG     Creation for version 1 only
**
************************************************************************/

/*

  Functions supporting condition variables and their attributes. 
  Condition variables block the q that the thread is running on 
  by waiting for an event.
  
		int pthread_cond_destroy
		int pthread_cond_init    
		int pthread_cond_wait    
		int pthread_cond_timedwait
		int pthread_cond_signal   
		int pthread_cond_broadcast
		int pthread_condattr_init 
		int pthread_condattr_destroy 
 */

#include <pthread.h>
#include "internal.h"
#include "stack.h"
#include "mutex.h"

/**************************************************************

  	function: int pthread_cond_destroy() 

	description:
 
  Destroys the condition variable.
 
 **************************************************************/
int pthread_cond_destroy(cond) 
pthread_cond_t 	* cond;
{
  if (cond == NO_COND) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  /*
   *   Check to see if anyone is one the queue waiting to be signalled
   *   If so we return an error.
   */
  
  if (cond->waiters) 
  {
    set_errno(EBUSY);
    return(-1);
  }

  /*
   * No one is waiting. Make the condition structure invalid
   * so future calls with this handle fail and then unlock it.
   */
  cond->flags     = FALSE;
  cond            = NO_COND;
  return(0);
}

/**************************************************************

  	function: int pthread_cond_init()

	description:
 
  Initializes the condition variable referred by cond with 
  attributes specified by attr. If attr is NULL, default values 
  are used.
 
 **************************************************************/
int pthread_cond_init(cond, attr)
pthread_cond_t 			* cond;
pthread_condattr_t 		* attr;
{
  if (cond == NO_COND) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  if (!attr) 
    attr = &pthread_condattr_default;

  cond->flags 	= attr->flags;
  cond->waiters = 0;
  cond->mutex 	= NO_MUTEX;

  pthread_enq_cond(cond);
  return(0);
}

/**************************************************************

  	function: int pthread_condattr_init(attr)

	description:
 		
  Initializes a condition attributes object with the default 
  values.
 
 
 **************************************************************/
int pthread_condattr_init(attr)
pthread_condattr_t      * attr;
{
  if (!attr) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  attr->flags = TRUE;
  return(0);
}

/**************************************************************

  	function: int pthread_condattr_destroy(attr)

	description:
 		
  Destoys a condition attribute object.
 
 **************************************************************/
int pthread_condattr_destroy(attr)
pthread_condattr_t 		*attr;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  attr->flags = FALSE;
  return(0);
}

/**************************************************************

  	function: int pthread_cond_wait(cond, mutex)

	description:
 		
  Atomically releases the mutex and causes the thread to block 
  on condition variable cond. The thread is taken off the thread 
  queues and queued onto condition variable with the state 
  changed to indicate the wait. The thread is then blocked 
  waiting for an event. This will be a signal event meaning that 
  the condition has been signalled or broadcast.  If it was 
  signalled the thread has already been reactivated and so the 
  mutex is reaquired and return.
 
 **************************************************************/
int pthread_cond_wait(cond, mutex)
pthread_cond_t 	            * cond;
pthread_mutex_t             * mutex;
{
  register tcb_t       * t;
  pthread_key_t        thiskey;
  
  thiskey = mac_pthread_key();
  t = &tcb[thiskey];
  
  if (cond == NO_COND || mutex == NO_MUTEX ||
      (cond->mutex != mutex && cond->mutex != NO_MUTEX) ||
      !(mutex->lock) || mutex->owner != thiskey) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  SET_KERNEL_FLAG();

  /*
   * clear error number before suspending
   */
  set_errno(0);

  cond->mutex 	= mutex;
  t->cond 		= cond;
  cond->waiters++;

  if (t->mutex != NULL)
  {
    t->mutex->lock = FALSE;
    t->mutex->owner = NO_OWNER;
    t->mutex = NULL;
  }
  
  mac_cond_wait(cond, thiskey);
  CONTEXT_SWITCH(TCB_WAITING);


  /*
   * relock the mutex under contention, but only if it was not already
   * locked as a side-effect of executing an signal handler and thereby
   * pthread_cond_wait_terminate(). Notice that locking the mutex before
   * executing the signal handler is a feature not required by the standard.
   */
  SET_KERNEL_FLAG();

  if (mutex->owner != thiskey) 
  {
    mac_mutex_lock(mutex, thiskey);
    if (!--(t->cond->waiters))
      cond->mutex = NO_MUTEX;
    t->cond = NO_COND;
  }
  
  CLEAR_KERNEL_FLAG();

  if (get_errno())
    return(-1);

  return(0);
}

/**************************************************************

  	function: int pthread_cond_timedwait()

	description:
 		
  This function works in the same way as pthread_cond_wait with 
  the exception of the added complexity of the timeout. The 
  timeout is delivered as a timeout event. If the timeout 
  happened then the thread is put back on the active queue.  It 
  will be implemented later on.
 
 **************************************************************/
int pthread_cond_timedwait(cond, mutex, timeout)
pthread_cond_t          * cond;
pthread_mutex_t         * mutex;
struct timespec         * timeout;
{
  register tcb_t       * t;
  pthread_key_t        thiskey;
  
  thiskey = mac_pthread_key();
  t = &tcb[thiskey];

  if (cond == NO_COND || mutex == NO_MUTEX ||
      (cond->mutex != mutex && cond->mutex != NO_MUTEX) ||
      !(mutex->lock) || mutex->owner != thiskey) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  if (timeout == NULL)
  {
    set_errno(EINVAL);
    return(-1);
  }
  
  set_errno(ENOSYS);
  return(-1);
}

/**************************************************************

  	function: void pthread_cond_wait_terminate()

	description:
 
  terminate conditional wait prematurely, relocks the mutex and 
  disassociates the condition variable (used when signal 
  received while in conditional wait) Notice: This is a feature. 
  Pthreads does not require that the mutex be locked inside a 
  user signal handler when a signal interrupts 
  pthread_cond_wait(). assumes SET_KERNEL_FLAG for 
  C_CONTEXT_SWITCH
 
 **************************************************************/
void pthread_cond_wait_terminate()
{
  register tcb_t       * t;
  pthread_key_t        thiskey;
  pthread_mutex_t      * mutex;
  
  thiskey = mac_pthread_key();
  t = &tcb[thiskey];

  SET_KERNEL_FLAG();

  /*
   * done only once by innermost user handler if handlers are nested
   * cond must therefore be retrieved out of TCB
   */
  if (t->cond != NO_COND) 
  {
    mutex = t->cond->mutex;
    mac_mutex_lock(mutex, thiskey);

    if (!--(t->cond->waiters))
      t->cond->mutex = NO_MUTEX;
    t->cond = NO_COND;
  }

  CLEAR_KERNEL_FLAG();
}

/**************************************************************

  	function: int pthread_cond_signal(cond)

	description:
 	
  Unblocks (by def. at least) one thread. Scan through the 
  waiter queue of threads looking for one. The first one found 
  is the very choice, and the TCB_READY will be the new state 
  for this thread. The mutex ownership is tranferred at once.
 
 **************************************************************/
int pthread_cond_signal(cond)
pthread_cond_t      * cond;
{
  if (cond == NO_COND) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  SET_KERNEL_FLAG();

  mac_cond_wakeup_one(cond);

  CLEAR_KERNEL_FLAG();
  return(0);
}

/**************************************************************

  	function: int pthread_cond_broadcast(cond)

	description:
 		
  Step through the queue of waiting threads and wakeup all the 
  threads.  The mutex ownership is tranferred right away.
 
 **************************************************************/
int pthread_cond_broadcast(cond)
pthread_cond_t      * cond;
{
  int               num;

  if (cond == NO_COND) 
  {
    set_errno(EINVAL);
    return(-1);
  }
  
  SET_KERNEL_FLAG();

  mac_cond_wakeup(cond, &num);
  
  CLEAR_KERNEL_FLAG();

  return(0);
}

