/******************************** SBB **********************************
**
** Project      :       Multiple-Threaded Library (MTL)
**
** Author       :       Zhang Xiangyang
**
** Date         :       November 4, 1996
**
** Version      :       1.0
**              :       Copyright (C) 1996, Zhang Xiangyang
**                
** File name    :       pthread.c
**
** Description  :
**
**      This file is part of pthread Library, which provides the thread 
**      interface compatible with POSIX 1003.4a Reference.
**
**      Report problems and direct all questions to:
**       zhangx@sebb.bel.alcatel.be
**
********************************HISTORY*********************************
**   DATE    INITIALS	COMMENTS		
**   ----    --------	--------
** 04/11/96  XZHANG     Creation for version 1 only
**
************************************************************************/

#ifndef _PTHREAD_INTERNAL_H_INCLUDED

#include <errno.h>
/*
 * Pthreads interface internals
 */
#define STAND_ALONE             
#define WINDOWSIZE              (0)
#define MINFRAME                (WINDOWSIZE + (7 * 4))
#define STACK_ALIGN             4
#define MAX_THREADS             100


/* Other Program Specific Constants */
#define MAX_PRIORITY            101
#define MIN_PRIORITY            0
#define DEFAULT_PRIORITY        MIN_PRIORITY
#define DEFAULT_STACKSIZE       0x1000
#define MAX_STACKSIZE           0x8000
#define PTHREAD_BODY_OFFSET     200
#define NO_PRIO                 MIN_PRIORITY 
#define NO_PRIO_INHERIT         101
#define PRIO_INHERIT            102 
#define PRIO_PROTECT            103

/*
 * stack, page alignment
 */
#define SA(X) (((X)+(STACK_ALIGN-1)) & ~(STACK_ALIGN-1))
#define PA(X) ((((int)X)+((int)pthread_page_size-1)) & \
				~((int)pthread_page_size-1))

#define TCB_FREE                0
#define TCB_USED                1
#define TCB_ALLOCATED           1
#define TCB_OCCUPIED            0

/*
 * Pthreads execution state in TCB and THREAD
 */
#define TCB_TERMINATED          0X0000
#define TCB_WAITING             0X0001
#define TCB_READY               0X0002
#define TCB_RUNNING             0X0004

#define T_MAIN                  0X0001
#define T_CONTROLLED            0X0002
#define T_DETACHED              0X0004
#define T_RETURNED              0X0008
#define T_EXECUTED              0X0010
#define T_SYNCTIMER             0X0020
#define T_CONDTIMER             0X0040
#define T_RUNNING               0X0080
#define T_INTR_POINT            0X0100

/* 
 * An array is defined in the kernel to represent the different 
 * kind of queues
 */
#define THREAD_QUEUE            0
#define MUTEX_QUEUE             1
#define COND_QUEUE              2
#define TIMER_QUEUE             3
#define NQUEUES                 6

/* 
 * structure describing queue element 
 * the node is defined universally only in order to simplify the 
 * problem.
 */
typedef struct queue
{
    int                 type;
    pthread_key_t       key;
    pthread_mutex_t     * mutex;
    pthread_cond_t      * cond;
    struct queue        * next;
    struct queue        * prev;
} pthread_queue_t;

/* 
 * thread control blocks are used to control the execution
 * of the threads internally.
 * For the moment, MTL has no capability of processing signals
 * so the jump buffer is jmpbuf, not sig_jmpbuf.
 * But it will be extended to include signal in the future.
 */
typedef struct 	TCB
{
    char                    allocated;         /* used for free opration */
    char                    used;
    int                     state;             /* thread execution state */
    pthread_key_t           key;
    char                    * oldsp, * newsp;
    jmp_buf                 context;           /* save area for context switch   */
    char                    * stack_base;      /* bottom of run-time stack       */
    char                    flag;              /* Automated                      */
    int                     status;            /* thread status                  */
    int                     osp, opc, obp;     /* save area for old context sw   */
    pthread_attr_t          attr;              /* attributes                     */
    int                     base_prio;         /* Base priority of thread        */
    int                     max_ceiling_prio;  /* Max of ceiling prio among locked mutexes*/
    int                     new_prio;          /* New Priority                            */
    pthread_func_t          func;              /* actual function to call upon activation */
    int                     arg;               /* argument list to above function         */
    int                     result;            /* return value of above function          */
    int                     cancel;
    int                     cancel_state;
    int                     specific;          /* It is a useless value */
    pthread_cond_t          * cond;            /* cond pointer if in cond_wait   */
    pthread_mutex_t         * mutex;           /* mutex pointer if in mutex_wait */
    pthread_key_t           joint;
    pthread_queue_t         * q_pos;           /* the entry point in queue       */
    /* the following are defined for future development */
    int                     wait_on_select;    /* more information in fds        */
    int                     width;             /* pertinent file desc. set width */
    int                     how_many;          /* how many amoung  0 .. width -1 */
    cleanup_t               cleanup_top;       /* stack of cleanup handlers      */
    fd_set                  readfds;           /* read file descriptor set       */
    fd_set                  writefds;          /* write file descriptor set      */
    fd_set                  exceptfds;         /* except. file descriptor set    */
    int                     sig;               /* latest received signal         */
    int                     code;              /* latest received signal code    */
    sigset_t                mask;              /* set of signals thread has masked out*/
    sigset_t                pending;           /* set of signals pending on thread    */
    sigset_t                sigwaitset;        /* set of signals thread is waiting for*/
    pthread_func_t          timer_func;        /* function to be called on timeout*/
    any_t                   timer_arg;         /* arg of above function           */
    int                     num_timers;        /* number of timers for thread     */
    struct timeval          interval;          /* time left for SCHED_RR          */
    struct timespec         tp;                /* wake-up time                    */
    int                     dummy;             /* filler for stack alignment      */
} tcb_t;


#define	MUTEX_INITIALIZER		{ 0, QUEUE_INITIALIZER, 0, 1 }
#define NO_MUTEX      			((pthread_mutex_t *)0)
#define MUTEX_VALID 			0x1
#define NO_OWNER				((int)(-1));


#define	CONDITION_INITIALIZER	{ QUEUE_INITIALIZER, 1, 0 }
#define NO_COND       			((pthread_cond_t *) 0)
#define COND_VALID 				0x1

#define	NO_PTHREAD				((pthread_t *) 0)
#define NO_ATTRIBUTE 			((pthread_attr_t *)0)

typedef struct kernel {
  pthread_key_t         k_pthread_key;     /* thread that is currently running   */
  volatile int 	        k_is_in_kernel;    /* flag to test if in kernel          */
  volatile int          k_state_change;    /* dispatcher state (run q/signals)   */
  char                  * k_process_stack_base;/* stack base of process          */
  pthread_queue_t       * k_qhead[NQUEUES];/* ready queue                        */
  pthread_queue_t       * k_qtail[NQUEUES];/* queue of all threads               */
  char                  * k_set_warning;   /* pointer to set warning message     */
  char                  * k_clear_warning; /* pointer to clear warning message   */
  char                  * k_prio_warning;  /* pointer to prio warning message    */
                                           /* for STAND_ALONE                    */
  volatile sigset_t     k_new_signals;     /* bit set of new signals to handle   */
  sigset_t              k_handlerset;      /* set of signals with user handler   */
  sigset_t              k_proc_mask;       /* Mask for process                   */
  int                   k_cur_heap;	   /* current break                      */
  sigset_t              k_pending_signals; /* bit set of pending signals         */
  sigset_t              k_all_signals;     /* mask of all (maskable) signals     */
  sigset_t              k_no_signals;      /* mask of no signals                 */
  sigset_t              k_cantmask;        /* mask of signals (cannot be caught) */
  int                   k_is_updating_timer;/* flag to test if timeout is handled*/
  volatile struct timespec k_timeofday;    /* Time of Day                        */
  int                   k_gwidth;          /* global width                       */
  fd_set                k_greadfds;        /* global read file descriptor set    */
  fd_set                k_gwritefds;       /* global write file descriptor set   */ 
  fd_set                k_gexceptfds;      /* global except file descriptor set  */ 
} kernel_t;

#ifdef PTHREAD_KERNEL_DEFINED
tcb_t                   tcb[MAX_THREADS];
kernel_t                pthread_kern;
pthread_condattr_t      pthread_condattr_default;
pthread_mutexattr_t     pthread_mutexattr_default;
pthread_attr_t          pthread_attr_default;

volatile int            pthread_started = FALSE;
volatile int            n_pthreads = 0;
volatile int            n_keys = 0;
#else
extern tcb_t                tcb[];
extern kernel_t             pthread_kern;
extern pthread_condattr_t   pthread_condattr_default;
extern pthread_mutexattr_t  pthread_mutexattr_default;
extern pthread_attr_t       pthread_attr_default;

extern volatile int         pthread_started;
extern volatile int         n_pthreads;
extern volatile int         n_keys;
#endif /* PTHREAD_KERNEL_DEFINED */

/* 		Internal Functions 
 *
 * changed for speed-up and interface purposes -
 * pthread_self() is now a function, 
 * but mac_pthread_self() is used internally
 */
#define mac_pthread_self() pthread_kern.k_pthread_key
#define mac_pthread_key()  pthread_kern.k_pthread_key
#define state_change       pthread_kern.k_state_change
#define is_in_kernel       pthread_kern.k_is_in_kernel
#define is_updating_timer  pthread_kern.k_is_updating_timer
#define new_signals        pthread_kern.k_new_signals
#define pending_signals    pthread_kern.k_pending_signals
#define all_signals        pthread_kern.k_all_signals
#define no_signals         pthread_kern.k_no_signals
#define cantmask           pthread_kern.k_cantmask
#define process_stack_base pthread_kern.k_process_stack_base
#define qhead              pthread_kern.k_qhead
#define qtail              pthread_kern.k_qtail
#define handlerset         pthread_kern.k_handlerset
#define set_warning        pthread_kern.k_set_warning
#define clear_warning      pthread_kern.k_clear_warning
#define prio_warning       pthread_kern.k_prio_warning

#define proc_mask          pthread_kern.k_proc_mask
#define cur_heap           pthread_kern.k_cur_heap
#define timeofday          pthread_kern.k_timeofday

#define gwidth             pthread_kern.k_gwidth    
#define greadfds           pthread_kern.k_greadfds  
#define gwritefds          pthread_kern.k_gwritefds 
#define gexceptfds         pthread_kern.k_gexceptfds

/* 		Internal Functions 
 *
 * using double-linked list to emulate queue
 * function concerning queue
 * 
 */
int pthread_enq(pthread_queue_t * node, int type);
int pthread_enq_thread(pthread_key_t key);
int pthread_enq_mutex(pthread_mutex_t * mutex);
int pthread_enq_cond(pthread_cond_t * cond);
int pthread_deq_thread(pthread_key_t key);
int pthread_deq_mutex(pthread_mutex_t * mutex);
int pthread_deq_cond(pthread_cond_t * cond);
int pthread_deq(pthread_queue_t * node, int type);
pthread_queue_t * pthread_queue_create(void);
pthread_queue_t * pthread_search_thread(pthread_key_t key);
pthread_queue_t * pthread_search_mutex(pthread_mutex_t * mutex);
pthread_queue_t * pthread_search_cond(pthread_cond_t * cond);


/* 		Internal Functions 
 *
 * using double-linked list to emulate queue
 * function concerning queue
 * 
 */
tcb_t * pthread_tcb_alloc(void);
int pthread_free(pthread_key_t key);
pthread_key_t pthread_scheduler(pthread_key_t key);
int mac_mutex_wait(pthread_mutex_t * mutex, pthread_key_t key);
int mac_mutex_wakeup(pthread_mutex_t * mutex, int * num);
int mac_cond_wait(pthread_cond_t * cond, pthread_key_t key);
int mac_cond_wakeup(pthread_cond_t * cond, int * num);
int mac_cond_wakeup_one(pthread_cond_t * cond);
void pthread_wakeup_join(void);
int mac_wakeup_join(pthread_key_t key);



#define MAX(x, y) ((x > y)? x : y)
#define MIN(x, y) ((x < y)? x : y)

/*
 * Errno is mapped on process' _errno 
 * and swapped upon context switch
 */
extern int errno;
#define set_errno(e)    (errno = (e))
#define get_errno()     (errno)

#define SET_KERNEL_FLAG()\
{\
    if (is_in_kernel)\
        fprintf(stderr, "WARNING : KERNEL = 1\n");\
    else\
        is_in_kernel = TRUE;\
}

#define CLEAR_KERNEL_FLAG()\
{\
    if (is_in_kernel)\
        is_in_kernel = FALSE;\
    else\
        fprintf(stderr, "WARNING : KERNEl = 0.\n");\
}


#define _PTHREAD_INTERNAL_H_INCLUDED
#endif /* _PTHREAD_INTERNAL_H_INCLUDED */

