/******************************** SBB **********************************
**
** Project      :       Multiple-Threaded Library (MTL)
**
** Author       :       Zhang Xiangyang
**
** Date         :       November 4, 1996
**
** Version      :       1.0
**              :       Copyright (C) 1996, Zhang Xiangyang
**                
** File name    :       pthread.c
**
** Description  :
**
**      This file is part of pthread Library, which provides the thread 
**      interface compatible with POSIX 1003.4a Reference.
**
**      Report problems and direct all questions to:
**       zhangx@sebb.bel.alcatel.be
**
********************************HISTORY*********************************
**   DATE    INITIALS	COMMENTS		
**   ----    --------	--------
** 04/11/96  XZHANG     Creation for version 1 only
**
************************************************************************/

/* 
 * Functions for the support of mutual exclusion - mutexes and their 
 * attributes.  
 
	int pthread_mutex_lock 
	int pthread_mutex_trylock 
	int pthread_mutex_unlock   
	int pthread_mutex_init    
	int pthread_mutex_destroy  
	int pthread_mutexattr_init 
	int pthread_mutexattr_destroy 
 */

#include <pthread.h>
#include "internal.h"
#include "stack.h"
#include "mutex.h"

/**************************************************************
 
  	function: int pthread_mutex_lock(mutex)

	description:

  Checks are made to see if the mutex is currently in use or 
  not.  If the mutex is not in use, then its locked, otherwise 
  the currently executing thread is put in the wait queue and a 
  new thread is selected. 
 
 **************************************************************/
int pthread_mutex_lock(mutex)
pthread_mutex_t        * mutex;
{
  register tcb_t       * t;
  pthread_key_t        thiskey;
  
  thiskey = mac_pthread_key();
  t = &tcb[thiskey];

  if (mutex == NO_MUTEX) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  if (mutex->owner == thiskey)
  {
    set_errno(EDEADLK);
    return(-1);
  }

  /*
   * check the owner and decide the operation
   *	
   * if locked, wait for mutex
   * if unlock, lock it and continue to execute
   */
  SET_KERNEL_FLAG();
  mac_mutex_lock(mutex, thiskey);
  CLEAR_KERNEL_FLAG();

  return(0);
}

/**************************************************************
 
  	function: int pthread_mutex_trylock(mutex)

	description:

  Just try to lock the mutex. If lock succeeds, it is similar to 
  pthread_mutex_lock(), otherwise return right away without 
  getting the lock. 
 
 **************************************************************/
int pthread_mutex_trylock(mutex)
pthread_mutex_t      *mutex;
{
  register tcb_t       * t;
  pthread_key_t        thiskey;
  
  thiskey = mac_pthread_key();
  t = &tcb[thiskey];

  if (mutex == NO_MUTEX) 
  {
    set_errno(EINVAL);
    return(-1);
  }

 
  /*
   * check the owner and decide the operation
   *	
   * if locked, return
   * if unlock, lock it and continue to execute
   */
  SET_KERNEL_FLAG();
  if (mutex->lock) 
  {
    CLEAR_KERNEL_FLAG();
    set_errno(EBUSY);
    return(-1);
  }
 
  mac_mutex_lock(mutex, thiskey);
 
  CLEAR_KERNEL_FLAG();
  return(0);
}

/**************************************************************
 
  	function: int pthread_mutex_unlock(mutex) 

	description:

  Called only by the owner of mutex to release the mutex, then 
  it will yield to let the other thread execute, which is not 
  fact defined in POSIX.
  
 **************************************************************/
int pthread_mutex_unlock(mutex)
pthread_mutex_t        * mutex;
{
  pthread_key_t        thiskey;
  
  thiskey = mac_pthread_key();

  if (mutex == NO_MUTEX || mutex->owner != thiskey) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  SET_KERNEL_FLAG();
  
  tcb[thiskey].mutex = NO_MUTEX;
  mac_mutex_unlock(mutex);
  
  CLEAR_KERNEL_FLAG();

  return(0);
}

/**************************************************************
 
  	function: int pthread_mutex_init(mutex, attr)

	description:

	Initialize the mutex and at the same time ensure that it's 
  usaable.  Set the attribute values from attr specified. No 
  check is made to see if the attributes are right as yet.
 
 **************************************************************/
int pthread_mutex_init(mutex, attr)
pthread_mutex_t         * mutex;
pthread_mutexattr_t     * attr;
{
  if (mutex == NO_MUTEX) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  if (!attr)
    attr = &pthread_mutexattr_default;

  mutex->owner          = NO_OWNER;
  mutex->flags          = attr->flags;
  mutex->prio_ceiling 	= attr->prio_ceiling;
  mutex->protocol       = attr->protocol;
  mutex->lock           = FALSE;
  
  pthread_enq_mutex(mutex);
  return(0);
}  

/**************************************************************
 
  	function: int pthread_mutex_destroy(mutex) 

	description:

  Destroys the mutex.
 
 **************************************************************/
int pthread_mutex_destroy(mutex) 
pthread_mutex_t     * mutex;
{
  if (mutex == NO_MUTEX) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  /*
   * free mutex only if not locked 
   * and not associated with any cond var
   */
  if (pthread_test_and_set(mutex)) 
  {
    set_errno(EBUSY);
    return(-1);
  }

  mutex->flags = FALSE;
  return(0);
}

/**************************************************************
 
  	function:   int pthread_test_and_set(mutex) 

	description:

  test if a mutex could be destroyed and return,

	0 : eligible
	1 : can not be destroyed
 
 **************************************************************/
int pthread_test_and_set(mutex) 
pthread_mutex_t *mutex;
{
  pthread_queue_t 		* q;

  /* check if mutex is being used */
  if (mutex->lock == TRUE)
    return (1);

  /* check if any cond uses the mutex */
  for (q = qhead[COND_QUEUE]; q != NULL; q = q->next)
  {
    if ((q->cond->waiters > 0) && (q->mutex == mutex))
      return (1);
  }
	
  return (0);
}

/**************************************************************
 
  	function: int pthread_mutex_setprio_ceiling(mutex, prio_ceiling)

	description:

  locks the mutex, changes the mutex's priority ceiling and 
  releases the mutex.
 
 **************************************************************/
int pthread_mutex_setprio_ceiling(mutex, prio_ceiling)
pthread_mutex_t 	* mutex;
int 				prio_ceiling;
{
  int 				oldprio;

  if (mutex == NO_MUTEX || !mutex->flags) 
  {
    set_errno(ESRCH);
    return(-1);
  }
  
  if (prio_ceiling >= MIN_PRIORITY && prio_ceiling <= MAX_PRIORITY) 
  {
    if (!pthread_mutex_trylock(mutex)) 
    {
      oldprio = mutex->prio_ceiling;
      mutex->prio_ceiling = prio_ceiling;
      pthread_mutex_unlock(mutex);
      return(oldprio);
    }
    else 
    {
      set_errno(EPERM);
      return(-1);
    }
  }
  else 
  {
    set_errno(EINVAL);
    return(-1);
  }
}

/**************************************************************
 
  	function: int pthread_mutex_getprio_ceiling(mutex)

	description:

  Returns the current priority ceiling of the mutex.
 
 **************************************************************/
int pthread_mutex_getprio_ceiling(mutex)
pthread_mutex_t mutex;
{
  if (!mutex.flags) 
  {
    set_errno(ESRCH);
    return(-1);
  }
  
  if (mutex.prio_ceiling >= MIN_PRIORITY &&
      mutex.prio_ceiling <= MAX_PRIORITY)
    return(mutex.prio_ceiling);
  else 
  {
    set_errno(EINVAL);
    return(-1);
  }
}

/**************************************************************
 
  	function: int pthread_mutexattr_init(attr)

	description:

  Initializes the mutex attribute object with default values.
 
 **************************************************************/
int pthread_mutexattr_init(attr)
pthread_mutexattr_t *attr;
{
  if (!attr)
  {
    set_errno(EINVAL);
    return(-1);
  }

  attr->flags           = TRUE;
  attr->prio_ceiling    = DEFAULT_PRIORITY;
  attr->protocol        = NO_PRIO_INHERIT;
  return(0);
}

/**************************************************************
  
  	function: int pthread_mutexattr_destroy(attr)

	description:

  Destroys the mutex attribute object.
 
 **************************************************************/
int pthread_mutexattr_destroy(attr)
pthread_mutexattr_t     *attr;
{

  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }
  
  attr->flags           = FALSE;
  attr->prio_ceiling    = DEFAULT_PRIORITY;
  attr->protocol        = NO_PRIO_INHERIT;
  return(0);
}

/**************************************************************
 
  	function: int pthread_mutexattr_setprotocol(attr,protocol)


	description:

  Sets the protocol (value can be NO_PRIO_INHERIT, PRIO_INHERIT 
  or PRIO_PROTECT) for the mutex attr.
 
 **************************************************************/
int pthread_mutexattr_setprotocol(attr,protocol)
pthread_mutexattr_t     * attr;
pthread_protocol_t      protocol;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  set_errno(ENOSYS);
  return(-1);

}

/**************************************************************
 
  	function: pthread_protocol_t pthread_mutexattr_getprotocol(attr)


	description:

  Gets the current protcol.
 
 
 **************************************************************/
pthread_protocol_t pthread_mutexattr_getprotocol(attr)
pthread_mutexattr_t 	attr;
{
  if (!attr.flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  set_errno(ENOSYS);
  return(-1);
}

/**************************************************************
 
  	function: int pthread_mutexattr_setprio_ceiling(attr,prio_ceiling)


	description:

  Sets the priority ceiling for the mutex attribute object.
 
 **************************************************************/
int pthread_mutexattr_setprio_ceiling(attr,prio_ceiling)
pthread_mutexattr_t     * attr;
int                     prio_ceiling;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  set_errno(ENOSYS);
  return(-1);
}

/**************************************************************
 
  	function: int pthread_mutexattr_getprio_ceiling(attr)

	description:

  Gets the current priority ceiling of the mutex attribute 
  object.
 
 **************************************************************/
int pthread_mutexattr_getprio_ceiling(attr)
pthread_mutexattr_t 	attr;
{
  if (!attr.flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  set_errno(ENOSYS);
  return(-1);
}

