/******************************** SBB **********************************
**
** Project      :       Multiple-Threaded Library (MTL)
**
** Author       :       Zhang Xiangyang
**
** Date         :       November 4, 1996
**
** Version      :       1.0
**              :       Copyright (C) 1996, Zhang Xiangyang
**                
** File name    :       pthread.c
**
** Description  :
**
**      This file is part of pthread Library, which provides the thread 
**      interface compatible with POSIX 1003.4a Reference.
**
**      Report problems and direct all questions to:
**       zhangx@sebb.bel.alcatel.be
**
********************************HISTORY*********************************
**   DATE    INITIALS	COMMENTS		
**   ----    --------	--------
** 04/11/96  XZHANG     Creation for version 1 only
**
************************************************************************/
/*
 * Implementation of fork, join, exit, etc.

 int 			pthread_create  
 int 			pthread_join     // block
 int 			pthread_detach
 void 			pthread_exit              
 void 			pthread_yield     
 pthread_t 		pthread_self
 int 			pthread_cancel
 int 			pthread_setcancel 
 void 			pthread_testcancel
 */

#include <stdlib.h>
#include <pthread.h>
#include "internal.h"
#include "stack.h"


/**************************************************************
 
  	function: 	void pthread_init(void)

	description:

		initialize the threads environment. This function is the 
 first function that must be called in any multi-threaded 
 program. It initializes the main thread and the thread control 
 blecks for all the threads to use. 
 
 **************************************************************/
void pthread_init()
{
  tcb_t              * tcon;
  int                i;

  if (pthread_started)
    return;

  /* 
   * initialize TCB
   */
  for (i = 0; i < MAX_THREADS; ++i)
  {
    tcb[i].key       = i;
    tcb[i].used      = TCB_FREE;
    tcb[i].allocated = TCB_OCCUPIED;
    tcb[i].state     = TCB_TERMINATED;
    tcb[i].stack_base = NULL;
    tcb[i].cond      = NO_COND;
    tcb[i].mutex     = NO_MUTEX;
    tcb[i].joint     = -1;
    tcb[i].cancel_state= CANCEL_ON;
    tcb[i].cancel    = FALSE;
  }

  pthread_mutexattr_init(&pthread_mutexattr_default);
  pthread_condattr_init(&pthread_condattr_default);
  pthread_attr_init(&pthread_attr_default);

  /*
   * allocate main thread
   */
  tcon = pthread_tcb_alloc();
	
  if (tcon == NULL) 
  {
    exit(-1);
  }
  
  setjmp(tcon->context);
  tcon->stack_base = (char *) tcon->context[JB_SP];
  tcon->state = TCB_RUNNING;

  tcon->func = (pthread_func_t) 0;
  tcon->arg = (any_t *) NULL;
  tcon->status |= T_MAIN | T_CONTROLLED;

  tcon->attr.prio 		= MIN_PRIORITY;
  tcon->base_prio 		= MIN_PRIORITY;
  tcon->max_ceiling_prio 	= MAX_PRIORITY;
  tcon->new_prio 		= MAX_PRIORITY;
  tcon->attr.sched		= SCHED_FIFO;

  /*
   * intialize kernel structure
   * and queue of existing thread
   */
  mac_pthread_key() = tcon->key;	
  pthread_enq_thread(tcon->key);
  state_change     	= FALSE;

  pthread_started       = TRUE;
  n_pthreads = 1; /* main thread counts, fd_server does not count */
}


/**************************************************************
 
  	function: pthread_t pthread_self()

	description:

 returns immutable thread ID of calling thread
 
 **************************************************************/
pthread_t pthread_self()
{
  return(mac_pthread_self());
}


/**************************************************************
 
  	function: pthread_terminate

	description:

 The thread will be terminated and destroyed, i.e. the thread 
 will not exist any more.  In future, cleanup handlers and 
 destructor functions will be called.
 
 **************************************************************/
static void pthread_terminate()
{
  register tcb_t                * t;
  register pthread_key_t        thiskey, nextkey;

  thiskey = mac_pthread_key();
  t = &tcb[thiskey];

  /*
   * dequeue thread and schedule someone else
   */

  SET_KERNEL_FLAG();

  /*
   * Terminating thread has to be detached if anyone tries to join
   * but the memory is not freed until the dispatcher is called.
   * This is required by pthread_join().
   * The result is copied into the TCB of the joining threads to
   * allow the memory of the current thread to be reclaimed before
   * the joining thread accesses the result.
   */
  mac_wakeup_join(t->key);

  /*
   * The last threads switches off the light and calls UNIX exit
   */
  --n_pthreads;
  if (!n_pthreads) 
  {
    exit(t->result);
  }
  
  /*
   * SCHEDULE other thread to run in the memory
   */
  t->state = TCB_TERMINATED;
  nextkey = pthread_scheduler(t->key);
  if (nextkey < 0)
    fprintf(stderr, "scheduler error.\n");

  tcb[nextkey].state = TCB_RUNNING;
  pthread_deq_thread(t->key);
  CLEAR_KERNEL_FLAG();

  mac_pthread_key() = nextkey;
  longjmp(tcb[nextkey].context, nextkey+1);
}

/**************************************************************
 
  function int pthread_create(t, attr, func, arg)
  
  description:
  
  Create a new thread of execution. The thread structure and a 
  queue will be created in the call, and they will be bound 
  together.  The completely created pthread will be executed at 
  once, which is different from other POSIX thread package.  The 
  key and stack for this thread are also assigned by the call.
  
 **************************************************************/
int pthread_create(thread, attr, func, arg)
pthread_t                   * thread;
pthread_attr_t              * attr;
pthread_func_t              func;
int                         arg;
{
  register tcb_t            * t, * parent_t;
  pthread_key_t             parent_key_t;
  
  parent_key_t = mac_pthread_key();
  parent_t = &tcb[parent_key_t];

  if (!attr)
    attr = &pthread_attr_default;

  if (!attr->flags || thread == NULL) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  t = pthread_tcb_alloc();

  if (t == NULL) 
  {
    set_errno(EAGAIN);
    return(-1);
  }

  t->func = func;
  t->arg = arg;
  t->state = T_RUNNING | T_CONTROLLED;
  
  t->max_ceiling_prio     = NO_PRIO;
  t->new_prio             = NO_PRIO;

  t->attr.stacksize       = attr->stacksize;
  t->attr.flags           = attr->flags;
  t->attr.contentionscope = parent_t->attr.contentionscope;
  t->attr.inheritsched    = parent_t->attr.inheritsched;
  t->attr.sched           = parent_t->attr.sched;
  t->attr.prio            = parent_t->base_prio;

  /*
   * initialize stack and queue for child thread
   * do context switch : save parent context
   *  run child thread
   *  change parent state TCB_READY
   * 
   */
  if (attr->detachstate)
    t->status |= T_DETACHED;

  SET_KERNEL_FLAG();

  ++n_pthreads; 
  pthread_stack_init(t->key);
  pthread_enq_thread(t->key);
  
  CLEAR_KERNEL_FLAG();

  parent_t->state = TCB_READY;
  mac_pthread_key() = t->key;	
    
  if (setjmp(parent_t->context) == 0)
  {
     SET_THREAD_CONTEXT(t->key);
     (t->func)(t->arg);
  }
  else
  {
    * thread = t->key;
  }

  return(0);
}


/**************************************************************
 
  function: tcb_t * pthread_tcb_alloc()

  description:

  allocate pthread TCB structure
 
 **************************************************************/
static tcb_t * pthread_tcb_alloc(void)
{
  register tcb_t            * t;
  int                       i;

  for (i = 0; i < MAX_THREADS; i++) 
  {
    t = (tcb_t *) (&tcb[i]);
    if (t->used == TCB_FREE) 
    {
      t->used = TCB_USED;
      break; 
    }
  }
  
  if (i >= MAX_THREADS)
    return(NULL);

  return(t);
}


/**************************************************************
 
  function: int pthread_detach()

  description:

  Detaching a running thread. If the thread has returned then 
  the resources are also freed.
 
 **************************************************************/
int pthread_detach(t)
pthread_t               * t;
{
  pthread_key_t         key;

  if (t == NULL)
  {
    set_errno(EINVAL);
    return(-1);
  }

  key = * t;

  SET_KERNEL_FLAG();

  if (tcb[key].status & T_DETACHED) 
  {
    CLEAR_KERNEL_FLAG();
    set_errno(ESRCH);
  }
  
  tcb[key].status |= T_DETACHED;

  pthread_free(key);

  CLEAR_KERNEL_FLAG();
  return(0);
}

/**************************************************************
 
  function: static int pthread_free(thread_ptr)

  description:

  free the resources allocated for a running thread.
 
 **************************************************************/
static int pthread_free(key)
pthread_key_t       key;
{
  register tcb_t    * t;


  if ((key < 0) || (key > MAX_THREADS))
  {
    set_errno(EINVAL);
    return(-1);
  }

  t = &tcb[key];

  SET_KERNEL_FLAG();

  if (t->state & T_DETACHED) 
  {
    CLEAR_KERNEL_FLAG();
    set_errno(ESRCH);
  }
  
  t->state |= T_DETACHED;

  pthread_stack_destroy(t->key);
  if (t->allocated == TCB_ALLOCATED) 
  {
    free(t);
    t = NULL;
  }

  CLEAR_KERNEL_FLAG();
  return(0);
}

/**************************************************************
 
  function: int pthread_join(thread, status)

  description:

  Wait for a thread to exit. If the status parameter is non-NULL 
  then that threads exit status is stored in it.
 
 **************************************************************/
int pthread_join(thread, status)
pthread_t           * thread;
int                 * status;
{
  register tcb_t    * t;
  pthread_key_t     thiskey;
  
  thiskey = mac_pthread_key();
  t = &tcb[thiskey];

  if (thread == NO_PTHREAD) 
  {
    set_errno(EINVAL);
    return(-1);
  }
  
  if (* thread == t->key) 
  {
    set_errno(EDEADLK);
    return(-1);
  }

  SET_KERNEL_FLAG();
  
  /*
   * clear error number before suspending
   */
  set_errno(0);
  CLEAR_KERNEL_FLAG();

  /*
   * this thread sleep until specified thread exit
   */
  if (tcb[* thread] != TCB_TERMINATED)
  {
    t->joint = * thread;
    CONTEXT_SWITCH(TCB_WAITING);
  }

  /*
   * status was copied into result field of current TCB by other thread
   */
  if (status)
    * status = tcb[*thread].result;
  return(0);
}


/**************************************************************
 
  function: void pthread_yield(void)

  description:

  Yield the processor to another thread.  The thread's state 
  will be TCB_READY but the prority will be the minimal.  That 
  is to say the thread will be executed again until other READY 
  thread has been executed once.
 
 **************************************************************/
void pthread_yield(void)
{
    /*
     * Place ourself at the end of the ready queue.
     * This allows the other ready processes to execute thus
     * in effect yielding the processor.
     */

    CONTEXT_SWITCH(TCB_READY);
}

/**************************************************************
 
  function: void pthread_exit(status)

  description:

  Save the exit status of the thread.  The other threads that 
  join with this thread will be TCB_READY.
 
 **************************************************************/
void pthread_exit(status)
int                 status;
{
  register tcb_t    * t;
  pthread_key_t     key;
  
  key = mac_pthread_key();
  t = &tcb[key];

  t->result = status;
  pthread_terminate();
}

/**************************************************************
 
  function: int pthread_stack_init(pthread_t * t)

  description:

  initialize the stack for the newly created thread.  The 
  function should be called first and only once.
 
 **************************************************************/
int pthread_stack_init(pthread_key_t key)
{
  register tcb_t    * t;
  unsigned char     * sp, * incsp;

  if ((key < 0) || (key > MAX_THREADS))
  {
    set_errno(EINVAL);
    return(-1);
  }

  t = &tcb[key];

  sp = (char *)calloc(t->attr.stacksize, sizeof(char));
  t->stack_base = sp;

  incsp = sp;
  incsp++;
  if (incsp > sp)
  {
    t->oldsp = sp + t->attr.stacksize;
  }
  else
  {
    t->oldsp = sp;
  }

  t->newsp = t->oldsp;
  return (0);
}

/**************************************************************
 
  function: int pthread_stack_destroy(pthread_t * t)

  description:

  free the resource for allocating the thread's stack. 
 
 **************************************************************/
static void pthread_stack_destroy(pthread_key_t key)
{
  char 		* ptr;

  ptr = tcb[key].stack_base;
  free(ptr);
  tcb[key].stack_base = NULL;
}

/**************************************************************
 
  function: int pthread_setcancel()

  description:
		
  enable or disable the current thread's general cancelability. 
  state:  CANCEL_ON / CANCEL_OFF
 
 **************************************************************/
int pthread_setcancel(int state)
{
  register tcb_t            * t;
  pthread_key_t             thiskey;
  
  if ((state != CANCEL_ON) && (state != CANCEL_OFF))
  {
    set_errno(EINVAL);
    return(-1);
  }

  thiskey = mac_pthread_key();
  t = &tcb[thiskey];

  t->cancel_state = state;
  return (0);
}

/**************************************************************
 
  function: void pthread_testcancel()

  description:

  request the delivery of a pending cancel to the current thread
 
 **************************************************************/
void pthread_testcancel(void)
{
  register tcb_t            * t;
  pthread_key_t             thiskey;

  thiskey = mac_pthread_key();
  t = &tcb[thiskey];
  
  /* here no request operation is required
   * only in accordance with the pthread_cancel,
   * because the cancel is delivery to the specified 
   * thread directly.
   * 
   * so remember that it is designed for stand alone mode.
   */
   
  if (t->cancel == TRUE)
    pthread_exit(t->result);
}


/**************************************************************
 
  function: int pthread_cancel(pthread_t * thread)

  description:

  allows a thread to request it or another thread terminate the 
  execution.
 
 **************************************************************/
int pthread_cancel(pthread_t * thread)
{
  if (thread == NULL)
  {
    set_errno(EINVAL);
    return(-1);
  }
  
  /*
   * Here the cancel should be arrange by a cancel queue.
   * the subroutine check the cancel flag.
   * if it is permitted, cancel can be delivered immediately
   * otherwise it will be enqueued.
   *
   * but in my library I just put it into that thread
   * only in order to simplify the problem
   *
   * attention should be paid to the situation that multiple
   * processor run concurrently.  This function
   * will not function correctly.  It is only on effect 
   * when in stand alone mode.
   */
   
  tcb[*thread].cancel = TRUE;
  return (0);
}



/**************************************************************
 
  function: int pthread_once(once_block, init_routine)

  description:

  The first call to this will call the init_routine() with no 
  arguments, otherwise it will return.
 
 **************************************************************/
int pthread_once(once_block, init_routine)
pthread_once_t          * once_block;
void                    (* init_routine)();
{
  if (!once_block->init) 
  {
    once_block->init = TRUE;
    pthread_mutex_init(&once_block->mutex, NULL);
  }
  else
  {
	return (-1);
  }

  pthread_mutex_lock(&once_block->mutex);

  if (!once_block->exec) 
  {
    once_block->exec = TRUE;
    (*init_routine)();
  }
  
  pthread_mutex_unlock(&once_block->mutex);
  return(0);
}

/**************************************************************
 
  function: int pthread_attr_init()

  description:

  Initializes the attr (thread attribute object)
  with the default values.
 
 **************************************************************/
int pthread_attr_init(attr)
pthread_attr_t                * attr;
{
  if (!attr) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  attr->flags           = TRUE;
  attr->contentionscope = PTHREAD_SCOPE_LOCAL;
  attr->inheritsched    = PTHREAD_DEFAULT_SCHED;
  attr->detachstate     = 0;
  attr->sched           = SCHED_FIFO;
  attr->stacksize       = DEFAULT_STACKSIZE;
  attr->prio            = DEFAULT_PRIORITY;
  
  return(0);
}

/**************************************************************
 
  function: int pthread_attr_destroy()

  description:

  Destroys the thread attribute object.
 
 **************************************************************/
int pthread_attr_destroy(attr)
pthread_attr_t                * attr;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }
  
  attr->flags = FALSE;
  return(0);
}


/**************************************************************
 
  function: int pthread_attr_setstacksize()

  description:

  sets the size of the stack to "stacksize" in thread attribute object.
  Aligns the "stacksize" to double word boundary.
 
 **************************************************************/
int pthread_attr_setstacksize(attr, stacksize)
pthread_attr_t          * attr;
size_t                  stacksize;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  attr->stacksize = SA(stacksize); /* stack align, see asm_linkage.h */
  return(0);    
}

/**************************************************************
 
  function: int pthread_attr_getstacksize()

  description:

  gets the stacksize from an pthread attribute object.
 
 **************************************************************/
int pthread_attr_getstacksize(attr, stacksize)
pthread_attr_t          * attr;
size_t                  * stacksize;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  * stacksize = attr->stacksize;    
  return(0);
}

/**************************************************************
 
  function: int pthread_attr_setscope()

  description:

  Set the contentionscope attribute in a thread attribute object.
 
 **************************************************************/
int pthread_attr_setscope(attr, contentionscope)
pthread_attr_t          * attr;
int                     contentionscope;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }
  
  if (contentionscope == PTHREAD_SCOPE_LOCAL) 
  {
    attr->contentionscope = contentionscope;
    return(0);    
  }
  else 
  {
    set_errno(EINVAL);
    return(-1);
  }
}

/**************************************************************
 
  function: int pthread_attr_setinheritsched()

  description:

  Set the inheritsched attribute.
 
 **************************************************************/
int pthread_attr_setinheritsched(attr, inherit)
pthread_attr_t          * attr;
int                     inherit;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }
  
  if (inherit == PTHREAD_INHERIT_SCHED || inherit == PTHREAD_DEFAULT_SCHED) 
  {
    attr->inheritsched = inherit;
    return(0);    
  }
  else 
  {
    set_errno(EINVAL);
    return(-1);
  }
}

/**************************************************************
 
  function: int pthread_attr_setscheduler()

  description:

  set the sched attribute
 
 **************************************************************/
int pthread_attr_setscheduler(t, scheduler)
pthread_t           * t;
int                 scheduler;
{
  register tcb_t    * tcon;
  
  
  if (t == NULL) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  tcon = &tcb[* t];
  if (scheduler == SCHED_FIFO) 
  {
    tcon->attr.sched = scheduler;
    return(0);    
  }
  else 
  {
    set_errno(EINVAL);
    return(-1);
  }
}

/**************************************************************
 
  function: int pthread_attr_getscheduler()

  description:

  get the sched attribute
 
 **************************************************************/
int pthread_attr_getscheduler(t)
pthread_t 		    * t;
{
  if (t == NULL) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  return(tcb[*t].attr.sched);    
}

/**************************************************************
 
  function: int pthread_attr_setprio()

  description:

  set the prio attribute
 
 **************************************************************/
int pthread_attr_setprio(attr, priority)
pthread_attr_t          * attr;
int                     priority;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  if (priority >= MIN_PRIORITY && priority <= MAX_PRIORITY) 
  {
    attr->prio = priority;
    return(0);    
  }
  else 
  {
    set_errno(EINVAL);
    return(-1);
  }
}

/**************************************************************
 
  function: int pthread_attr_getscope()

  description:

  get the contentionscope attribute
 
 **************************************************************/
int pthread_attr_getscope(attr)
pthread_attr_t      * attr;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  return(attr->contentionscope);
}

/**************************************************************
 
  function: int pthread_attr_getinheritsched()

  description:

  get the inheritsched attribute
 
 **************************************************************/
int pthread_attr_getinheritsched(attr)
pthread_attr_t      * attr;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }
  return(attr->inheritsched);
}

/**************************************************************
 
  function: int pthread_attr_getsched()

  description:

  get the sched attribute
 
 **************************************************************/
int pthread_attr_getsched(attr)
pthread_attr_t      * attr;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  return(attr->sched);
}

/**************************************************************
 
  function: int pthread_attr_getprio()

  description:

  get the prio attribute
 
 **************************************************************/
int pthread_attr_getprio(attr)
pthread_attr_t 	*attr;
{
  if (!attr || !attr->flags) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  return(attr->prio);
}


/**************************************************************
 
  function: int pthread_attr_setdetachstate()

  description:

  Sets the detachstate attribute in  attr object
 
 **************************************************************/
int pthread_attr_setdetachstate(attr, detachstate)
pthread_attr_t 	    * attr;
int                 detachstate;
{
  if (!attr || !attr->flags || (detachstate > 1) || (detachstate < 0)) 
  {
    set_errno(EINVAL);
    return(-1);
  }
  attr->detachstate = detachstate;
  return(0);
}

/**************************************************************
 
  function: int pthread_attr_getdetachstate()

  description:

  Gets the value of detachstate attribute from attr object
 
 **************************************************************/
int pthread_attr_getdetachstate(attr)
pthread_attr_t 	* attr;
{
  if (!attr || !attr->flags) 
  { 
    set_errno(EINVAL);
    return(-1);    
  } 
  return(attr->detachstate); 
}



/**************************************************************
 
  function: int pthread_get_priority_max()

  description:

  inquire maximum priority value
 
 **************************************************************/
int pthread_get_priority_max()
{
  return(MAX_PRIORITY);
}

/**************************************************************
 
  function: int pthread_get_priority_min()

  description:

  inquire minimum priority value
 
 **************************************************************/
int pthread_get_priority_min()
{
  return(MIN_PRIORITY);
}



             /*  Attention */
      /* The following function will not available */
      /* in the version 1, donot use them now */


/**************************************************************
 
  	function: int pthread_key_create()

	description:

	   creates a new global key and spefies destructor call. 
	   returns -1 upon error with errno ENOMEM if name space 
	   exhausted and EAGAIN if insufficient memory.
 
 **************************************************************/
int pthread_key_create(key, destructor)
pthread_key_t   * key;
void            (* destructor)();
{
  if ((key == NULL) || (destructor == NULL))
  {
    set_errno(EINVAL);
    return(-1);
  }

  set_errno(ENOSYS);
  return(-1);
}



/**************************************************************
 
  	function: int pthread_setspecific()

	description:

  associate a value with a data key for some thread
  returns -1 upon error with errno EINVAL if the key is invalid.
 
 **************************************************************/
int pthread_setspecific(key, value)
pthread_key_t   key;
int             value;
{
  if (key < 0 || key >= n_keys) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  tcb[key].specific = value;
  return(0);
}

/**************************************************************
 
  	function: int pthread_getspecific()

	description:

  retrieve a value from a data key for some thread
  returns -1 upon error with errno EINVAL if the key is invalid.
 
 **************************************************************/
int pthread_getspecific(key, value)
pthread_key_t       key;
int                 * value;
{
  if (key < 0) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  * value = tcb[key].specific;
  return(0);
}

/**************************************************************
 
  	function: int pthread_cleanup_push()

	description:

  push function on current thread's cleanup stack
  and execute it with the specified argument.
 
 **************************************************************/
int pthread_cleanup_push(func, arg)
void                (*func)();
any_t               arg;
{
  register tcb_t    * t;
  pthread_key_t     thiskey;
  cleanup_t         new;

  thiskey = mac_pthread_key();
  t = &tcb[thiskey];

  if (!func) 
  {
    set_errno(EINVAL);
    return(-1);
  }

  if (!(new = (cleanup_t) malloc(sizeof(*new)))) 
  {
    set_errno(ENOMEM);
    return(-1);
  }

  new->func = func;
  new->arg = arg;

  new->next = t->cleanup_top;
  t->cleanup_top = new;

  return(0);
}


/**************************************************************
 
  	function: int pthread_cleanup_pop()

	description:

  pop function off current thread's cleanup stack.
  If execute is non-zero, execute it.
 
 **************************************************************/
int pthread_cleanup_pop(execute)
int                 execute;
{
  register tcb_t    * t;
  pthread_key_t     thiskey;
  cleanup_t         new;

  thiskey = mac_pthread_key();
  t = &tcb[thiskey];

  if (!(new = t->cleanup_top)) 
  {
    set_errno(EINVAL);
    return(-1);
  }
  
  t->cleanup_top = new->next;
  
  if (execute)
    (new->func)(new->arg);
  free(new);

  return(0);
}


