/******************************** SBB **********************************
**
** Project      :       Multiple-Threaded Library (MTL)
**
** Author       :       Zhang Xiangyang
**
** Date         :       November 4, 1996
**
** Version      :       1.0
**              :       Copyright (C) 1996, Zhang Xiangyang
**                
** File name    :       pthread.c
**
** Description  :
**
**      This file is part of pthread Library, which provides the thread 
**      interface compatible with POSIX 1003.4a Reference.
**
**      Report problems and direct all questions to:
**       zhangx@sebb.bel.alcatel.be
**
********************************HISTORY*********************************
**   DATE    INITIALS	COMMENTS		
**   ----    --------	--------
** 04/11/96  XZHANG     Creation for version 1 only
**
************************************************************************/

/*
 * Implementation of double-linked list emulating queues
 *
 * different data type can be represent with one void * 
 * type universally, but here I use different data type
 * only in order to make thread library work correctly
 * It can be updated in new version of thread library.
 */

#include <stdlib.h>
#include <pthread.h>
#include "internal.h"

int      n_queues = 0;

/**************************************************************
 
 	function : pthread_enq( ) 
	
	description :
	
	enqueing a node into an unsorted 
    queue (double-linked list)
 
 **************************************************************/
int pthread_enq(node, type)
pthread_queue_t         * node;
int                     type;
{
  pthread_queue_t       * q;

  if ( node == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  if ((type < 0) || (type >= NQUEUES))
  {
    set_errno(EINVAL);
    return (-1);
  }

  if ((qhead[type] == NULL) || (qtail[type] == NULL))
  {
    qhead[type] = qtail[type] = node;
  }
  else
  { 
    q = qtail[type];
    q->next = node;
    node->prev = q;
    qtail[type] = node;
  }
  
  return(0);
}

/**************************************************************
 
 	function : pthread_enq_thread( ) 
	
	description :
	
	enqueing a pthread into an unsorted 
    queue (linked list)
 
 **************************************************************/
int pthread_enq_thread(key)
pthread_key_t           key;
{
  pthread_queue_t       * cur;

  if ((key < 0) || (key > MAX_THREADS))
  {
    set_errno(EINVAL);
    return (-1);
  }

  cur = pthread_queue_create();
  
  if (cur == NULL)
  {
    set_errno(ENOMEM);
    return (-1);
  }
  
  cur->type = THREAD_QUEUE;
  cur->key = key;  
  return (pthread_enq(cur, THREAD_QUEUE));
}

/**************************************************************
 
 	function : pthread_enq_mutex( )
	
	description :
	
	enqueing a mutex into an unsorted 
    queue (doubled-linked list)
 
 **************************************************************/
int pthread_enq_mutex(mutex)
pthread_mutex_t         * mutex;
{
  pthread_queue_t       * cur;

  if (mutex == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  cur = pthread_queue_create();
  
  if (cur == NULL)
  {
    set_errno(ENOMEM);
    return (-1);
  }
  
  cur->type = MUTEX_QUEUE;
  cur->mutex = mutex;
  return (pthread_enq(cur, MUTEX_QUEUE));
}

/**************************************************************
 
 	function : pthread_enq_cond( ) 
	
	description :
	
	enqueing a cond into an unsorted 
    queue (doubled-linked list)
 
 **************************************************************/
int pthread_enq_cond(cond)
pthread_cond_t          * cond;
{
  pthread_queue_t       * cur;

  if (cond == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  cur = pthread_queue_create();
  
  if (cur == NULL)
  {
	set_errno(ENOMEM);
 	return (-1);
  }
  
  cur->type = COND_QUEUE;
  cur->cond = cond;
  return (pthread_enq(cur, COND_QUEUE));
}

/**************************************************************
 
 	function :  pthread_queue_t * pthread_queue_create( );
	
	description :
 
 **************************************************************/
pthread_queue_t * pthread_queue_create(void)
{
  pthread_queue_t       * q;
  
  q = (pthread_queue_t *) calloc(1, sizeof(pthread_queue_t));
  if (q == NULL)
  {
    set_errno(ENOMEM);
    return (q);
  }
  
  q->next = NULL;
  q->prev = NULL;
 
  n_queues ++; 
  return (q);
}

/**************************************************************
 
 	function : pthread_deq_thread() 
	
	description :
	
	- dequeue a thread from a queue
 
 **************************************************************/
int pthread_deq_thread(key)
pthread_key_t           key;
{
  pthread_queue_t       * cur;

  if ((key < 0) || (key > MAX_THREADS))
  {
    set_errno(EINVAL);
    return (-1);
  }

  cur = pthread_search_thread(key);
  return(pthread_deq(cur, THREAD_QUEUE));
}

/**************************************************************
 
 	function : pthread_deq_mutex() 
	
	description :
	
	- dequeue a mutex from a queue
 
 **************************************************************/
int pthread_deq_mutex(mutex)
pthread_mutex_t         * mutex;
{
  pthread_queue_t       * cur;

  if (mutex == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  cur = pthread_search_mutex(mutex);
  return(pthread_deq(cur, MUTEX_QUEUE));
}

/**************************************************************
 
 	function : pthread_deq_cond()
	
	description :
	
	- dequeue a thread from a queue
 
 **************************************************************/
int pthread_deq_cond(cond)
pthread_cond_t          * cond;
{
  pthread_queue_t       * cur;

  if (cond == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  cur = pthread_search_cond(cond);
  return(pthread_deq(cur, COND_QUEUE));
}



/**************************************************************
 
 	function : pthread_deq * pthread_deq( );
	
	description :
	
 
 **************************************************************/
int pthread_deq(node, type)
pthread_queue_t         * node;
int                     type;
{

 if (node == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  if ((type < 0) || (type >= NQUEUES))
  {
    set_errno(EINVAL);
    return (-1);
  }

  if ((qhead[type] == NULL) || (qtail[type] == NULL))
  {
    set_errno(EINVAL);
    return (-1);
  }

  if ((node == qhead[type]) || (node == qtail[type]))
  {
    if ((node == qhead[type]) && (node == qtail[type]))
    {
      qhead[type] = qtail[type] = NULL;
    }
    else if (node == qhead[type]) 
    {
      qhead[type] = node->next;
      node->next->prev = NULL;
    }
    else
    {
      qtail[type] = node->prev;
      node->prev->next = NULL;
    }
  }
  else
  {
    node->next->prev = node->prev;
    node->prev->next = node->next; 
  }

  free(node);
  n_queues --;

  return (0);
}


/**************************************************************
 
 	function :  pthread_queue_t * pthread_search_thread();
	
	description :
 
 **************************************************************/
pthread_queue_t * pthread_search_thread(key)
pthread_key_t           key;
{
  pthread_queue_t       * q;

  if ((key < 0) || (key > MAX_THREADS))
  {
    set_errno(EINVAL);
    return (NULL);
  }

  for (q = qhead[THREAD_QUEUE]; q != NULL;)
  {
    if (q->key == key)
    {
      return(q);
    }
    else
    {
      q = q->next;
    }
  } 
  
  set_errno(1);
  return (NULL);
}

/**************************************************************
 
 	function :  pthread_queue_t * pthread_search_mutex();
	
	description :
 
 **************************************************************/
pthread_queue_t * pthread_search_mutex(mutex)
pthread_mutex_t         * mutex;
{
  pthread_queue_t       * q;

  if (mutex == NULL)
  {
    set_errno(EINVAL);
    return (NULL);
  }

  for (q = qhead[MUTEX_QUEUE]; q != NULL;)
  {
    if (q->mutex == mutex)
    {
      return(q);
    }
    else
    {
      q = q->next;
    }
  } 
  
  set_errno(1);
  return (NULL);
}

/**************************************************************
 
 	function :  pthread_queue_t * pthread_search_cond( );
	
	description :
 
 **************************************************************/
pthread_queue_t * pthread_search_cond(cond)
pthread_cond_t          * cond;
{
  pthread_queue_t       * q;

  if (cond == NULL)
  {
    set_errno(EINVAL);
    return (NULL);
  }

  for (q = qhead[COND_QUEUE]; q != NULL;)
  {
    if (q->cond == cond)
    {
      return(q);
    }
    else
    {
      q = q->next;
    }
  } 
  
  set_errno(1);
  return (NULL);
}
