/******************************** SBB **********************************
**
** Project      :       Multiple-Threaded Library (MTL)
**
** Author       :       Zhang Xiangyang
**
** Date         :       November 4, 1996
**
** Version      :       1.0
**              :       Copyright (C) 1996, Zhang Xiangyang
**                
** File name    :       pthread.c
**
** Description  :
**
**      This file is part of pthread Library, which provides the thread 
**      interface compatible with POSIX 1003.4a Reference.
**
**      Report problems and direct all questions to:
**       zhangx@sebb.bel.alcatel.be
**
********************************HISTORY*********************************
**   DATE    INITIALS	COMMENTS		
**   ----    --------	--------
** 04/11/96  XZHANG     Creation for version 1 only
**
************************************************************************/

/*
 * Thread schedule
 */

#include <pthread.h>

#define PTHREAD_KERNEL_DEFINED
#include "internal.h"
#undef PTHREAD_KERNEL_DEFINED

/**************************************************************
 
  function : int pthread_scheduler(key)
  
  description : (FIFO scheduler)
	
	this function is to find the thread which will be scheduled to
	be next execution.  It checks the thread state in the queue and tried 
	to find the eligible thread.  If nothing is found, i.e. the 
	deadlock happens, it will terminated the thread program.  
	This function constitute the base of the pthreads execution.
	It works as a thread scheduler.
 
 **************************************************************/

pthread_key_t pthread_scheduler(pthread_key_t key)
{
  int                   found;
  pthread_key_t         newkey;
  pthread_queue_t       * q;
  
  
  q = pthread_search_thread(key);
  
  found = FALSE;
  for (q = q->next; q != NULL; q = q->next)
  {
    newkey = q->key;
	
    if (tcb[newkey].state == TCB_READY)
    {
      found = TRUE;
      break;
    }
  }
  
  if (found == FALSE)
  {
    for (q = qhead[THREAD_QUEUE]; q != NULL; q = q->next)
    {
      newkey = q->key;
	
      if (tcb[newkey].state == TCB_READY)
      {
        found = TRUE;
        break;
      }
    }
  }

  if (found == TRUE)
    return (newkey);
  else
    return(-1);
}


/**************************************************************
 
 	function : int mac_mutex_wait(mutex, t) 
	
	description :
	
	wait for the mutex unlocked.
 
 **************************************************************/

int mac_mutex_wait(mutex, key)
pthread_mutex_t         * mutex;
pthread_key_t           key;
{
  if ((key < 0) || (key > MAX_THREADS))
  {
    set_errno(EINVAL);
    return (-1);
  }

  if (mutex == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  /*
   * this thread is waiting for the given mutex
   */
  tcb[key].state = TCB_WAITING;
  tcb[key].mutex = mutex;
  return (0);
}

/**************************************************************
 
 	function : int mac_mutex_wakeup(mutex, &num) 
	
	description :
	
	enqueing a node into an unsorted 
    queue (double-linked list)
 
 **************************************************************/

int mac_mutex_wakeup(mutex, num)
pthread_mutex_t             * mutex;
int                         * num;
{
  pthread_queue_t           * q;

  if (mutex == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  * num = 0;

  for (q = qhead[THREAD_QUEUE]; q != NULL; q = q->next)
  {
    pthread_key_t 	thiskey;

    thiskey = q->key;
    if (tcb[thiskey].state == TCB_WAITING)
    {
      if (tcb[thiskey].mutex == mutex)
      {
        tcb[thiskey].state = TCB_READY;
        (*num)++;
        tcb[thiskey].mutex = NULL;
      }
    }
  }
  
  return (0);
}

/**************************************************************
 
 	function : int mac_cond_wait(cond, t) 
	
	description :
	
	wait for the cond unlocked.
 
 **************************************************************/

int mac_cond_wait(cond, key)
pthread_cond_t          * cond;
pthread_key_t           key;
{
  if ((key < 0) || (key > MAX_THREADS))
  {
    set_errno(EINVAL);
    return (-1);
  }

  if (cond == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  /*
   * this thread is waiting for the given cond
   */
  tcb[key].state = TCB_WAITING;
  tcb[key].cond = cond;
  return (0);
}

/**************************************************************
 
 	function : int mac_cond_wakeup(cond, &num) 
	
	description :
	
	enqueing a node into an unsorted 
    queue (double-linked list)
 
 **************************************************************/

int mac_cond_wakeup(cond, num)
pthread_cond_t              * cond;
int                         * num;
{
  pthread_queue_t           * q;

  if (cond == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  * num = 0;

  for (q = qhead[THREAD_QUEUE]; q != NULL; q = q->next)
  {
    pthread_key_t       thiskey;

    thiskey = q->key;
    if (tcb[thiskey].state == TCB_WAITING)
    {
      if (tcb[thiskey].cond == cond)
      {
        tcb[thiskey].state = TCB_READY;
        (*num)++;
        tcb[thiskey].cond = NULL;
      }
    }
  }
  
  return (0);
}

/**************************************************************
 
 	function : int mac_cond_wakeup_one() 
	
	description :
	
	enqueing a node into an unsorted 
    queue (double-linked list)
 
 **************************************************************/

int mac_cond_wakeup_one(cond)
pthread_cond_t              * cond;
{
  pthread_queue_t           * q;

  if (cond == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  for (q = qhead[THREAD_QUEUE]; q != NULL; q = q->next)
  {
    pthread_key_t       thiskey;

    thiskey = q->key;
    if (tcb[thiskey].state == TCB_WAITING)
    {
      if (tcb[thiskey].cond == cond)
      {
        tcb[thiskey].state = TCB_READY;
        tcb[thiskey].cond = NULL;
        return (0);
      }
    }
  }
  
  return (0);
}

/**************************************************************
 
 	function : int mac_wakeup_join() 
	
	description :
	
	enqueing a node into an unsorted 
    queue (double-linked list)
 
 **************************************************************/

int mac_wakeup_join(key)
pthread_key_t          key;
{
  pthread_queue_t      * q;

  if ((key < 0) || (key > MAX_THREADS))
  {
    set_errno(EINVAL);
    return (-1);
  }

  for (q = qhead[THREAD_QUEUE]; q != NULL; q = q->next)
  {
    pthread_key_t 	thiskey;

    thiskey = q->key;
    if (tcb[thiskey].state == TCB_WAITING)
    {
      if (tcb[thiskey].joint == key)
      {
        tcb[thiskey].state = TCB_READY;
        tcb[thiskey].joint = -1;
      }
    }
  }
  
  return (0);
}

/**************************************************************
 
 	function : int mac_thread_ready(int * num) 
	
	description :
	
	enqueing a node into an unsorted 
    queue (double-linked list)
 
 **************************************************************/

int mac_thread_ready(int * num)
{
  pthread_queue_t      * q;

  if (num == NULL)
  {
    set_errno(EINVAL);
    return (-1);
  }

  * num = 0;
  for (q = qhead[THREAD_QUEUE]; q != NULL; q = q->next)
  {
    pthread_key_t 	thiskey;

    thiskey = q->key;
    if (tcb[thiskey].state == TCB_READY)
    {
      (*num)++;
    }
  }
  
  return (0);
}

